/*
 * Name:    mp_modbus.c
 *
 * Purpose: Python module interface to generic MX MODBUS functions.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_modbus.h"
#include "mx_handle.h"
#include "mp.h"

MX_EXPORT PyObject *
MpPrivate_modbus_read_coils_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "read_coils";

	PyObject *coil_value_tuple, *coil_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_coils;
	mx_uint8_type *coil_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &starting_address, &num_coils );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	coil_value_array = (mx_uint8_type *)
			malloc( num_coils * sizeof(mx_uint8_type) );

	if ( coil_value_array == (mx_uint8_type *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_coils );
		return NULL;
	}

	mx_status = mx_modbus_read_coils( record,
					starting_address, num_coils,
					coil_value_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( coil_value_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	coil_value_tuple = PyTuple_New( (int) num_coils );

	if ( coil_value_tuple == NULL ) {
		mx_free( coil_value_array );
		return NULL;
	}

	for ( i = 0; i < num_coils; i++ ) {

		coil_value_object = PyInt_FromLong((long) coil_value_array[i]);

		if ( coil_value_object == NULL ) {
			Py_DECREF( coil_value_tuple );
			mx_free( coil_value_array );
			return NULL;
		}

		PyTuple_SET_ITEM( coil_value_tuple, i, coil_value_object );
	}

	mx_free( coil_value_array );

	return coil_value_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_discrete_inputs_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "read_discrete_inputs";

	PyObject *input_value_tuple, *input_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_inputs;
	mx_uint8_type *input_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &starting_address, &num_inputs );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	input_value_array = (mx_uint8_type *)
			malloc( num_inputs * sizeof(mx_uint8_type) );

	if ( input_value_array == (mx_uint8_type *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_inputs );
		return NULL;
	}

	mx_status = mx_modbus_read_discrete_inputs( record,
					starting_address, num_inputs,
					input_value_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( input_value_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	input_value_tuple = PyTuple_New( (int) num_inputs );

	if ( input_value_tuple == NULL ) {
		mx_free( input_value_array );
		return NULL;
	}

	for ( i = 0; i < num_inputs; i++ ) {

		input_value_object =
			PyInt_FromLong((long) input_value_array[i]);

		if ( input_value_object == NULL ) {
			Py_DECREF( input_value_tuple );
			mx_free( input_value_array );
			return NULL;
		}

		PyTuple_SET_ITEM( input_value_tuple, i, input_value_object );
	}

	mx_free( input_value_array );

	return input_value_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_holding_registers_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "read_holding_registers";

	PyObject *register_value_tuple, *register_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_registers;
	mx_uint16_type *register_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &starting_address, &num_registers );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	register_value_array = (mx_uint16_type *)
			malloc( num_registers * sizeof(mx_uint16_type) );

	if ( register_value_array == (mx_uint16_type *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_registers );
		return NULL;
	}

	mx_status = mx_modbus_read_holding_registers( record,
					starting_address, num_registers,
					register_value_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( register_value_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	register_value_tuple = PyTuple_New( (int) num_registers );

	if ( register_value_tuple == NULL ) {
		mx_free( register_value_array );
		return NULL;
	}

	for ( i = 0; i < num_registers; i++ ) {

		register_value_object =
			PyInt_FromLong((long) register_value_array[i]);

		if ( register_value_object == NULL ) {
			Py_DECREF( register_value_tuple );
			mx_free( register_value_array );
			return NULL;
		}

		PyTuple_SET_ITEM( register_value_tuple, i,
					register_value_object );
	}

	mx_free( register_value_array );

	return register_value_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_input_registers_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "read_input_registers";

	PyObject *register_value_tuple, *register_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_registers;
	mx_uint16_type *register_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &starting_address, &num_registers );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	register_value_array = (mx_uint16_type *)
			malloc( num_registers * sizeof(mx_uint16_type) );

	if ( register_value_array == (mx_uint16_type *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_registers );
		return NULL;
	}

	mx_status = mx_modbus_read_input_registers( record,
					starting_address, num_registers,
					register_value_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( register_value_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	register_value_tuple = PyTuple_New( (int) num_registers );

	if ( register_value_tuple == NULL ) {
		mx_free( register_value_array );
		return NULL;
	}

	for ( i = 0; i < num_registers; i++ ) {

		register_value_object =
			PyInt_FromLong((long) register_value_array[i]);

		if ( register_value_object == NULL ) {
			Py_DECREF( register_value_tuple );
			mx_free( register_value_array );
			return NULL;
		}

		PyTuple_SET_ITEM( register_value_tuple, i,
					register_value_object );
	}

	mx_free( register_value_array );

	return register_value_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_write_single_coil_cmd( PyObject *self, PyObject *args )
{
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int output_address, coil_value;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &output_address, &coil_value );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_modbus_write_single_coil( record,
						output_address, coil_value );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_modbus_write_single_register_cmd( PyObject *self, PyObject *args )
{
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int output_address, int_register_value;
	mx_uint16_type register_value;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
				&handle, &output_address, &int_register_value );

	if ( parse_status == 0 )
		return NULL;

	register_value = (mx_uint16_type) int_register_value;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_modbus_write_single_register( record,
					output_address, register_value );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_exception_status_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	MX_RECORD *record;
	long handle;
	int parse_status;
	mx_uint8_type exception_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "l", &handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_modbus_read_exception_status( record,
						&exception_status );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	result = Py_BuildValue( "i", (int) exception_status );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_modbus_get_comm_event_counter_cmd( PyObject *self, PyObject *args )
{
	PyObject *value_tuple, *value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	mx_uint16_type status_word, event_count;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "l", &handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_modbus_get_comm_event_counter( record,
						&status_word, &event_count );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Build a tuple to contain the returned values. */

	value_tuple = PyTuple_New( 2 );

	if ( value_tuple == NULL )
		return NULL;

	value_object = PyInt_FromLong( (long) status_word );

	if ( value_object == NULL ) {
		Py_DECREF( value_tuple );
		return NULL;
	}

	PyTuple_SET_ITEM( value_tuple, 0, value_object );

	value_object = PyInt_FromLong( (long) event_count );

	if ( value_object == NULL ) {
		Py_DECREF( value_tuple );
		return NULL;
	}

	PyTuple_SET_ITEM( value_tuple, 1, value_object );

	return value_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_write_multiple_coils_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "write_multiple_coils";

	PyObject *coil_value_tuple, *coil_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_coils;
	mx_uint8_type *coil_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "liO", &handle,
				&starting_address, &coil_value_tuple );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* The last object passed to us should be a tuple. */

	if ( PyTuple_Check( coil_value_tuple ) == 0 )
		return NULL;

	/* Find out how many values are in the tuple. */

	num_coils = (unsigned int) PyTuple_Size( coil_value_tuple );

	if ( num_coils == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
		"%s: The coil value tuple passed has the illegal size of 0.",
			cname );
		return NULL;
	}

	/* Allocate an array to hold the values we are going to send
	 * to the MODBUS interface.
	 */

	coil_value_array = (mx_uint8_type *)
			malloc( num_coils * sizeof(mx_uint8_type) );

	if ( coil_value_array == NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_coils );
		return NULL;
	}

	/* Copy the values from the tuple to the array. */

	for ( i = 0; i < num_coils; i++ ) {

		coil_value_object = PyTuple_GetItem( coil_value_tuple, i );

		if ( coil_value_object == NULL ) {
			mx_free( coil_value_array );
			return NULL;
		}

		if ( PyInt_Check( coil_value_object ) == 0 ) {
			mx_free( coil_value_array );
			return NULL;
		}

		coil_value_array[i] =
			(mx_uint8_type) PyInt_AsLong( coil_value_object );
	}

	/* Now we can write the values to the MODBUS interface. */

	mx_status = mx_modbus_write_multiple_coils( record,
						starting_address, num_coils,
						coil_value_array );

	mx_free( coil_value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_modbus_write_multiple_registers_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "write_multiple_registers";

	PyObject *register_value_tuple, *register_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, starting_address, num_registers;
	mx_uint16_type *register_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "liO", &handle,
				&starting_address, &register_value_tuple );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* The last object passed to us should be a tuple. */

	if ( PyTuple_Check( register_value_tuple ) == 0 )
		return NULL;

	/* Find out how many values are in the tuple. */

	num_registers = (unsigned int) PyTuple_Size( register_value_tuple );

	if ( num_registers == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
	"%s: The register value tuple passed has the illegal size of 0.",
			cname );
		return NULL;
	}

	/* Allocate an array to hold the values we are going to send
	 * to the MODBUS interface.
	 */

	register_value_array = (mx_uint16_type *)
			malloc( num_registers * sizeof(mx_uint16_type) );

	if ( register_value_array == NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, num_registers );
		return NULL;
	}

	/* Copy the values from the tuple to the array. */

	for ( i = 0; i < num_registers; i++ ) {

		register_value_object =
				PyTuple_GetItem( register_value_tuple, i );

		if ( register_value_object == NULL ) {
			mx_free( register_value_array );
			return NULL;
		}

		if ( PyInt_Check( register_value_object ) == 0 ) {
			mx_free( register_value_array );
			return NULL;
		}

		register_value_array[i] =
			(mx_uint16_type) PyInt_AsLong( register_value_object );
	}

	/* Now we can write the values to the MODBUS interface. */

	mx_status = mx_modbus_write_multiple_registers( record,
						starting_address, num_registers,
						register_value_array );

	mx_free( register_value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_modbus_mask_write_register_cmd( PyObject *self, PyObject *args )
{
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int register_address, int_and_mask, int_or_mask;
	mx_uint16_type and_mask, or_mask;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "liii", &handle,
			&register_address, &int_and_mask, &int_or_mask );

	if ( parse_status == 0 )
		return NULL;

	and_mask = (mx_uint16_type) int_and_mask;

	or_mask = (mx_uint16_type) int_or_mask;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_modbus_mask_write_register( record,
					register_address, and_mask, or_mask );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_write_multiple_registers_cmd( PyObject *self,
							PyObject *args )
{
	static const char cname[] = "read_write_multiple_registers";

	PyObject *read_register_tuple, *read_register_object;
	PyObject *write_register_tuple, *write_register_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, read_starting_address, num_registers_to_read;
	unsigned int write_starting_address, num_registers_to_write;
	mx_uint16_type *read_register_array, *write_register_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "liiiO", &handle,
				&read_starting_address, &num_registers_to_read,
				&write_starting_address, &write_register_tuple);

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* The last object passed to us should be a tuple. */

	if ( PyTuple_Check( write_register_tuple ) == 0 )
		return NULL;

	/* Find out how many values are in the tuple. */

	num_registers_to_write =
		(unsigned int) PyTuple_Size( write_register_tuple );

	if ( num_registers_to_write == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
	"%s: The write register tuple passed has the illegal size of 0.",
			cname );
		return NULL;
	}

	/* Allocate an array to hold the values we are going to write
	 * to the MODBUS interface.
	 */

	write_register_array = (mx_uint16_type *)
		malloc( num_registers_to_write * sizeof(mx_uint16_type) );

	if ( write_register_array == NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
"%s: Ran out of memory trying to allocate a %u element write register array.",
			cname, num_registers_to_write );
		return NULL;
	}

	/* Allocate another array to hold the values we are going to read
	 * from the MODBUS interface.
	 */

	read_register_array = (mx_uint16_type *)
		malloc( num_registers_to_read * sizeof(mx_uint16_type) );

	if ( read_register_array == NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
"%s: Ran out of memory trying to allocate a %u element read register array.",
			cname, num_registers_to_read );
		mx_free( write_register_array );
		return NULL;
	}

	/* Copy the values from the write register tuple to the write array. */

	for ( i = 0; i < num_registers_to_write; i++ ) {

		write_register_object =
				PyTuple_GetItem( write_register_tuple, i );

		if ( write_register_object == NULL ) {
			mx_free( write_register_array );
			mx_free( read_register_array );
			return NULL;
		}

		if ( PyInt_Check( write_register_object ) == 0 ) {
			mx_free( write_register_array );
			mx_free( read_register_array );
			return NULL;
		}

		write_register_array[i] =
			(mx_uint16_type) PyInt_AsLong( write_register_object );
	}

	/* Communicate with the MODBUS interface. */

	mx_status = mx_modbus_read_write_multiple_registers( record,
						read_starting_address,
						num_registers_to_read,
						read_register_array,
						write_starting_address,
						num_registers_to_write,
						write_register_array );

	mx_free( write_register_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( read_register_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	read_register_tuple = PyTuple_New( (int) num_registers_to_read );

	if ( read_register_tuple == NULL ) {
		mx_free( read_register_array );
		return NULL;
	}

	for ( i = 0; i < num_registers_to_read; i++ ) {

		read_register_object =
			PyInt_FromLong((long) read_register_array[i]);

		if ( read_register_object == NULL ) {
			Py_DECREF( read_register_tuple );
			mx_free( read_register_array );
			return NULL;
		}

		PyTuple_SET_ITEM( read_register_tuple, i,
					read_register_object );
	}

	mx_free( read_register_array );

	return read_register_tuple;
}

MX_EXPORT PyObject *
MpPrivate_modbus_read_fifo_queue_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "read_fifo_queue";

	PyObject *fifo_value_tuple, *fifo_value_object;
	MX_RECORD *record;
	long handle;
	int parse_status;
	unsigned int i, fifo_pointer_address;
	unsigned int max_fifo_values, num_fifo_values_read;
	mx_uint16_type *fifo_value_array;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lii",
			&handle, &fifo_pointer_address, &max_fifo_values );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	fifo_value_array = (mx_uint16_type *)
			malloc( max_fifo_values * sizeof(mx_uint16_type) );

	if ( fifo_value_array == (mx_uint16_type *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
	"%s: Ran out of memory trying to allocate a %u element array.",
			cname, max_fifo_values );
		return NULL;
	}

	mx_status = mx_modbus_read_fifo_queue( record,
					fifo_pointer_address,
					max_fifo_values,
					fifo_value_array,
					&num_fifo_values_read );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( fifo_value_array );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Build a tuple to contain the returned values. */

	fifo_value_tuple = PyTuple_New( (int) num_fifo_values_read );

	if ( fifo_value_tuple == NULL ) {
		mx_free( fifo_value_array );
		return NULL;
	}

	for ( i = 0; i < num_fifo_values_read; i++ ) {

		fifo_value_object =
			PyInt_FromLong((long) fifo_value_array[i]);

		if ( fifo_value_object == NULL ) {
			Py_DECREF( fifo_value_tuple );
			mx_free( fifo_value_array );
			return NULL;
		}

		PyTuple_SET_ITEM( fifo_value_tuple, i,
					fifo_value_object );
	}

	mx_free( fifo_value_array );

	return fifo_value_tuple;
}

